package fr.username404.snowygui.config

import com.typesafe.config.*
import com.typesafe.config.ConfigFactory.*
import fr.username404.snowygui.ClickGui
import fr.username404.snowygui.Snowy
import fr.username404.snowygui.gui.elements.ClickBox
import fr.username404.snowygui.gui.feature.Category
import fr.username404.snowygui.gui.feature.Macro
import io.github.config4k.extract
import io.github.config4k.getValue
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.serialization.decodeFromString
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import net.minecraft.client.Minecraft
import java.io.File
import kotlin.reflect.KProperty

object Configuration {
    private val ModifiableValues: MutableMap<String, () -> ConfigValue> = mapOf(
        "enabledFeatures" to { ConfigValueFactory.fromMap(enabledFeatures) },
        "macros" to { ConfigValueFactory.fromIterable((ClickGui.components.find { it is ClickBox && it.isCategory(Category.MACROS) } as ClickBox).buttons.map {
            Json.encodeToString(it as Macro)
        }) },
        "box_colors" to { ConfigValueFactory.fromAnyRef(mapOf<String, Int>(*ClickGui.clickboxes.filter { it.name != null }.map { it.name!!.key to it.color }.toTypedArray())) },
    ).toMutableMap()
    private fun Config.withFullModifiableValues() = ModifiableValues.entries.fold(this) { previous, entry ->
        previous.withValue(entry.key, entry.value())
    }
    private val configDirectory: String = (Minecraft.getInstance().gameDirectory.absolutePath + File.separator + "config").also { File(it).mkdir() }
    private val file: File = File(configDirectory + File.separator + "snowy.conf")
    private suspend fun writeConfig(c: Config) = coroutineScope {
        launch {
            file.writeText(
                """
                |Snowy {
                |${
                      c.root().render(
                          ConfigRenderOptions.defaults()
                              .setFormatted(true)
                              .setJson(false)
                              .setOriginComments(false)
                      ).prependIndent("  ").trimEnd()
                  }
                |}
                """.trimMargin()
            )
        }
    }

    operator fun getValue(ref: Any?, property: KProperty<*>): Boolean = invoke().getValue(ref, property)
    operator fun <T> setValue(ref: Any?, property: KProperty<*>, value: T) { ModifiableValues[property.name] = { ConfigValueFactory.fromAnyRef(value) } }

    operator fun invoke() = obtained

    private val obtained: Config by lazy {
        var result: Config = empty()
        with(file) {
            if (!exists()) {
                createNewFile()
                setWritable(true)
                setReadable(true)
            } else try {
                result = parseFile(file)
            } catch (e: ConfigException) {
                Snowy.logs.warn("Could not parse the snowy configuration file, the default configuration will be used instead.")
            }
        }
        load(result).withFallback(
            parseString(
                """
                    |Snowy {
                    |  displayInitMessage = false
                    |  macros = []
                    |  riskyCheats = false
                    |  sortAlphabetically = true
                    |}
                """.trimMargin()
            )
        ).extract<Config>("Snowy")
    }.also {
        runBlocking {
            writeConfig(it.value)
        }
    }
    val enabledFeatures = mutableMapOf<String, Boolean>().apply {
        if (obtained.hasPath("enabledFeatures")) {
            putAll(obtained.extract("enabledFeatures"))
        }
    }
    val foundMacros: Set<Macro> = run {
        obtained.getStringList("macros").map {
            Json.decodeFromString<Macro>(it)
        }.toSet()
    }
    init {
        Runtime.getRuntime().addShutdownHook(
            Thread {
                runBlocking {
                    ClickGui.boxContext {
                        enabledFeatures.putAll(buttons.map {
                            it.title to it.toggled
                        })
                    }
                    writeConfig(obtained.withFullModifiableValues()).join()
                }
            }
        )
    }
}
